@@ -1,32 +1,70 @@
 # -*- coding: utf-8 -*-
-from __future__ import unicode_literals, print_function
+from __future__ import print_function, unicode_literals
 
 import logging
 import os
 import re
+from collections import OrderedDict
 
 import docutils
 import docutils.core
 import docutils.io
-from docutils.writers.html4css1 import HTMLTranslator
+from docutils.writers.html4css1 import HTMLTranslator, Writer
+
 import six
+from six import StringIO
+from six.moves.html_parser import HTMLParser
 
-# import the directives to have pygments support
 from pelican import rstdirectives  # NOQA
+from pelican import signals
+from pelican.cache import FileStampDataCacher
+from pelican.contents import Author, Category, Page, Tag
+from pelican.utils import SafeDatetime, escape_html, get_date, pelican_open, \
+    posixize_path
+
 try:
     from markdown import Markdown
 except ImportError:
     Markdown = False  # NOQA
-try:
-    from html import escape
-except ImportError:
-    from cgi import escape
-from six.moves.html_parser import HTMLParser
 
-from pelican import signals
-from pelican.cache import FileStampDataCacher
-from pelican.contents import Page, Category, Tag, Author
-from pelican.utils import get_date, pelican_open, SafeDatetime, posixize_path
+# Metadata processors have no way to discard an unwanted value, so we have
+# them return this value instead to signal that it should be discarded later.
+# This means that _filter_discardable_metadata() must be called on processed
+# metadata dicts before use, to remove the items with the special value.
+_DISCARD = object()
+
+DUPLICATES_DEFINITIONS_ALLOWED = {
+    'tags': False,
+    'date': False,
+    'modified': False,
+    'status': False,
+    'category': False,
+    'author': False,
+    'save_as': False,
+    'url': False,
+    'authors': False,
+    'slug': False
+}
+
+METADATA_PROCESSORS = {
+    'tags': lambda x, y: ([
+        Tag(tag, y)
+        for tag in ensure_metadata_list(x)
+    ] or _DISCARD),
+    'date': lambda x, y: get_date(x.replace('_', ' ')),
+    'modified': lambda x, y: get_date(x),
+    'status': lambda x, y: x.strip() or _DISCARD,
+    'category': lambda x, y: _process_if_nonempty(Category, x, y),
+    'author': lambda x, y: _process_if_nonempty(Author, x, y),
+    'authors': lambda x, y: ([
+        Author(author, y)
+        for author in ensure_metadata_list(x)
+    ] or _DISCARD),
+    'slug': lambda x, y: x.strip() or _DISCARD,
+}
+
+logger = logging.getLogger(__name__)
+
 
 def ensure_metadata_list(text):
     """Canonicalize the format of a list of authors or tags.  This works
@@ -46,14 +84,9 @@ def ensure_metadata_list(text):
         else:
             text = text.split(',')
 
-    return [v for v in (w.strip() for w in text) if v]
-
-
-# Metadata processors have no way to discard an unwanted value, so we have
-# them return this value instead to signal that it should be discarded later.
-# This means that _filter_discardable_metadata() must be called on processed
-# metadata dicts before use, to remove the items with the special value.
-_DISCARD = object()
+    return list(OrderedDict.fromkeys(
+        [v for v in (w.strip() for w in text) if v]
+    ))
 
 
 def _process_if_nonempty(processor, name, settings):
@@ -64,28 +97,11 @@ def _process_if_nonempty(processor, name, settings):
     return processor(name, settings) if name else _DISCARD
 
 
-METADATA_PROCESSORS = {
-    'tags': lambda x, y: ([Tag(tag, y) for tag in ensure_metadata_list(x)]
-                          or _DISCARD),
-    'date': lambda x, y: get_date(x.replace('_', ' ')),
-    'modified': lambda x, y: get_date(x),
-    'status': lambda x, y: x.strip() or _DISCARD,
-    'category': lambda x, y: _process_if_nonempty(Category, x, y),
-    'author': lambda x, y: _process_if_nonempty(Author, x, y),
-    'authors': lambda x, y: ([Author(author, y)
-                              for author in ensure_metadata_list(x)]
-                             or _DISCARD),
-    'slug': lambda x, y: x.strip() or _DISCARD,
-}
-
-
 def _filter_discardable_metadata(metadata):
     """Return a copy of a dict, minus any items marked as discardable."""
     return {name: val for name, val in metadata.items() if val is not _DISCARD}
 
 
-logger = logging.getLogger(__name__)
-
 class BaseReader(object):
     """Base class to read files.
 
@@ -134,12 +150,19 @@ class _FieldBodyTranslator(HTMLTranslator):
         pass
 
 
-def render_node_to_html(document, node):
-    visitor = _FieldBodyTranslator(document)
+def render_node_to_html(document, node, field_body_translator_class):
+    visitor = field_body_translator_class(document)
     node.walkabout(visitor)
     return visitor.astext()
 
 
+class PelicanHTMLWriter(Writer):
+
+    def __init__(self):
+        Writer.__init__(self)
+        self.translator_class = PelicanHTMLTranslator
+
+
 class PelicanHTMLTranslator(HTMLTranslator):
 
     def visit_abbreviation(self, node):
@@ -159,11 +182,26 @@ class PelicanHTMLTranslator(HTMLTranslator):
 
 
 class RstReader(BaseReader):
-    """Reader for reStructuredText files"""
+    """Reader for reStructuredText files
+
+    By default the output HTML is written using
+    docutils.writers.html4css1.Writer and translated using a subclass of
+    docutils.writers.html4css1.HTMLTranslator. If you want to override it with
+    your own writer/translator (e.g. a HTML5-based one), pass your classes to
+    these two attributes. Look in the source code for details.
+
+        writer_class                    Used for writing contents
+        field_body_translator_class     Used for translating metadata such
+            as article summary
+
+    """
 
     enabled = bool(docutils)
     file_extensions = ['rst']
 
+    writer_class = PelicanHTMLWriter
+    field_body_translator_class = _FieldBodyTranslator
+
     class FileInput(docutils.io.FileInput):
         """Patch docutils.io.FileInput to remove "U" mode in py3.
 
@@ -180,18 +218,27 @@ class RstReader(BaseReader):
     def __init__(self, *args, **kwargs):
         super(RstReader, self).__init__(*args, **kwargs)
 
-    def _parse_metadata(self, document):
+    def _parse_metadata(self, document, source_path):
         """Return the dict containing document metadata"""
         formatted_fields = self.settings['FORMATTED_FIELDS']
 
         output = {}
+
+        if document.first_child_matching_class(docutils.nodes.title) is None:
+            logger.warning(
+                'Document title missing in file %s: '
+                'Ensure exactly one top level section',
+                source_path)
+
         for docinfo in document.traverse(docutils.nodes.docinfo):
             for element in docinfo.children:
                 if element.tagname == 'field':  # custom fields (e.g. summary)
                     name_elem, body_elem = element.children
                     name = name_elem.astext()
                     if name in formatted_fields:
-                        value = render_node_to_html(document, body_elem)
+                        value = render_node_to_html(
+                            document, body_elem,
+                            self.field_body_translator_class)
                     else:
                         value = body_elem.astext()
                 elif element.tagname == 'authors':  # author list
@@ -209,20 +256,23 @@ class RstReader(BaseReader):
         extra_params = {'initial_header_level': '2',
                         'syntax_highlight': 'short',
                         'input_encoding': 'utf-8',
-                        'exit_status_level': 2,
+                        'language_code': self.settings.get('DEFAULT_LANG'),
+                        'halt_level': 2,
+                        'traceback': True,
+                        'warning_stream': StringIO(),
                         'embed_stylesheet': False}
         user_params = self.settings.get('DOCUTILS_SETTINGS')
         if user_params:
             extra_params.update(user_params)
 
         pub = docutils.core.Publisher(
+            writer=self.writer_class(),
             source_class=self.FileInput,
             destination_class=docutils.io.StringOutput)
         pub.set_components('standalone', 'restructuredtext', 'html')
-        pub.writer.translator_class = PelicanHTMLTranslator
         pub.process_programmatic_settings(None, extra_params, None)
         pub.set_source(source_path=source_path)
-        pub.publish(enable_exit_status=True)
+        pub.publish()
         return pub
 
     def read(self, source_path):
@@ -231,7 +281,7 @@ class RstReader(BaseReader):
         parts = pub.writer.parts
         content = parts.get('body')
 
-        metadata = self._parse_metadata(pub.document)
+        metadata = self._parse_metadata(pub.document, source_path)
         metadata.setdefault('title', parts.get('title'))
 
         return content, metadata
@@ -245,9 +295,14 @@ class MarkdownReader(BaseReader):
 
     def __init__(self, *args, **kwargs):
         super(MarkdownReader, self).__init__(*args, **kwargs)
-        self.extensions = list(self.settings['MD_EXTENSIONS'])
-        if 'meta' not in self.extensions:
-            self.extensions.append('meta')
+        settings = self.settings['MARKDOWN']
+        settings.setdefault('extension_configs', {})
+        settings.setdefault('extensions', [])
+        for extension in settings['extension_configs'].keys():
+            if extension not in settings['extensions']:
+                settings['extensions'].append(extension)
+        if 'markdown.extensions.meta' not in settings['extensions']:
+            settings['extensions'].append('markdown.extensions.meta')
         self._source_path = None
 
     def _parse_metadata(self, meta):
@@ -258,17 +313,18 @@ class MarkdownReader(BaseReader):
         for name, value in meta.items():
             name = name.lower()
             if name in formatted_fields:
-                # handle summary metadata as markdown
-                # summary metadata is special case and join all list values
-                summary_values = "\n".join(value)
+                # formatted metadata is special case and join all list values
+                formatted_values = "\n".join(value)
                 # reset the markdown instance to clear any state
                 self._md.reset()
-                summary = self._md.convert(summary_values)
-                output[name] = self.process_metadata(name, summary)
-            elif name in METADATA_PROCESSORS:
+                formatted = self._md.convert(formatted_values)
+                output[name] = self.process_metadata(name, formatted)
+            elif not DUPLICATES_DEFINITIONS_ALLOWED.get(name, True):
                 if len(value) > 1:
-                    logger.warning('Duplicate definition of `%s` '
-                        'for %s. Using first one.', name, self._source_path)
+                    logger.warning(
+                        'Duplicate definition of `%s` '
+                        'for %s. Using first one.',
+                        name, self._source_path)
                 output[name] = self.process_metadata(name, value[0])
             elif len(value) > 1:
                 # handle list metadata as list of string
@@ -282,11 +338,14 @@ class MarkdownReader(BaseReader):
         """Parse content and metadata of markdown files"""
 
         self._source_path = source_path
-        self._md = Markdown(extensions=self.extensions)
+        self._md = Markdown(**self.settings['MARKDOWN'])
         with pelican_open(source_path) as text:
             content = self._md.convert(text)
 
-        metadata = self._parse_metadata(self._md.Meta)
+        if hasattr(self._md, 'Meta'):
+            metadata = self._parse_metadata(self._md.Meta)
+        else:
+            metadata = {}
         return content, metadata
 
 
@@ -347,7 +406,7 @@ class HTMLReader(BaseReader):
                 self._in_body = False
                 self._in_top_level = True
             elif self._in_body:
-                self._data_buffer += '</{}>'.format(escape(tag))
+                self._data_buffer += '</{}>'.format(escape_html(tag))
 
         def handle_startendtag(self, tag, attrs):
             if tag == 'meta' and self._in_head:
@@ -368,11 +427,16 @@ class HTMLReader(BaseReader):
             self._data_buffer += '&#{};'.format(data)
 
         def build_tag(self, tag, attrs, close_tag):
-            result = '<{}'.format(escape(tag))
+            result = '<{}'.format(escape_html(tag))
             for k, v in attrs:
-                result += ' ' + escape(k)
+                result += ' ' + escape_html(k)
                 if v is not None:
-                    result += '="{}"'.format(escape(v))
+                    # If the attribute value contains a double quote, surround
+                    # with single quotes, otherwise use double quotes.
+                    if '"' in v:
+                        result += "='{}'".format(escape_html(v, quote=False))
+                    else:
+                        result += '="{}"'.format(escape_html(v, quote=False))
             if close_tag:
                 return result + ' />'
             return result + '>'
@@ -380,7 +444,8 @@ class HTMLReader(BaseReader):
         def _handle_meta_tag(self, attrs):
             name = self._attr_value(attrs, 'name')
             if name is None:
-                attr_serialized = ', '.join(['{}="{}"'.format(k, v) for k, v in attrs])
+                attr_list = ['{}="{}"'.format(k, v) for k, v in attrs]
+                attr_serialized = ', '.join(attr_list)
                 logger.warning("Meta tag in file %s does not have a 'name' "
                                "attribute, skipping. Attributes: %s",
                                self._filename, attr_serialized)
@@ -394,13 +459,23 @@ class HTMLReader(BaseReader):
                         "Meta tag attribute 'contents' used in file %s, should"
                         " be changed to 'content'",
                         self._filename,
-                        extra={'limit_msg': ("Other files have meta tag "
-                                             "attribute 'contents' that should "
-                                             "be changed to 'content'")})
+                        extra={'limit_msg': "Other files have meta tag "
+                                            "attribute 'contents' that should "
+                                            "be changed to 'content'"})
 
             if name == 'keywords':
                 name = 'tags'
-            self.metadata[name] = contents
+
+            if name in self.metadata:
+                # if this metadata already exists (i.e. a previous tag with the
+                # same name has already been specified then either convert to
+                # list or append to list
+                if isinstance(self.metadata[name], list):
+                    self.metadata[name].append(contents)
+                else:
+                    self.metadata[name] = [self.metadata[name], contents]
+            else:
+                self.metadata[name] = contents
 
         @classmethod
         def _attr_value(cls, attrs, name, default=None):
@@ -474,7 +549,8 @@ class Readers(FileStampDataCacher):
 
         path = os.path.abspath(os.path.join(base_path, path))
         source_path = posixize_path(os.path.relpath(path, base_path))
-        logger.debug('Read file %s -> %s',
+        logger.debug(
+            'Read file %s -> %s',
             source_path, content_class.__name__)
 
         if not fmt:
@@ -486,7 +562,8 @@ class Readers(FileStampDataCacher):
                 'Pelican does not know how to parse %s', path)
 
         if preread_signal:
-            logger.debug('Signal %s.send(%s)',
+            logger.debug(
+                'Signal %s.send(%s)',
                 preread_signal.name, preread_sender)
             preread_signal.send(preread_sender)
 
@@ -527,20 +604,26 @@ class Readers(FileStampDataCacher):
             def typogrify_wrapper(text):
                 """Ensures ignore_tags feature is backward compatible"""
                 try:
-                    return typogrify(text, self.settings['TYPOGRIFY_IGNORE_TAGS'])
+                    return typogrify(
+                        text,
+                        self.settings['TYPOGRIFY_IGNORE_TAGS'])
                 except TypeError:
                     return typogrify(text)
 
             if content:
                 content = typogrify_wrapper(content)
+
+            if 'title' in metadata:
                 metadata['title'] = typogrify_wrapper(metadata['title'])
 
             if 'summary' in metadata:
                 metadata['summary'] = typogrify_wrapper(metadata['summary'])
 
         if context_signal:
-            logger.debug('Signal %s.send(%s, <metadata>)',
-                context_signal.name, context_sender)
+            logger.debug(
+                'Signal %s.send(%s, <metadata>)',
+                context_signal.name,
+                context_sender)
             context_signal.send(context_sender, metadata=metadata)
 
         return content_class(content=content, metadata=metadata,
@@ -560,7 +643,7 @@ def find_empty_alt(content, path):
             # src before alt
             <img
             [^\>]*
-            src=(['"])(.*)\1
+            src=(['"])(.*?)\1
             [^\>]*
             alt=(['"])\3
         )|(?:
@@ -569,7 +652,7 @@ def find_empty_alt(content, path):
             [^\>]*
             alt=(['"])\4
             [^\>]*
-            src=(['"])(.*)\5
+            src=(['"])(.*?)\5
         )
         """, re.X)
     for match in re.findall(imgs, content):
@@ -591,8 +674,12 @@ def default_metadata(settings=None, process=None):
             if process:
                 value = process('category', value)
             metadata['category'] = value
-        if settings.get('DEFAULT_DATE', None) and settings['DEFAULT_DATE'] != 'fs':
-            metadata['date'] = SafeDatetime(*settings['DEFAULT_DATE'])
+        if settings.get('DEFAULT_DATE', None) and \
+           settings['DEFAULT_DATE'] != 'fs':
+            if isinstance(settings['DEFAULT_DATE'], six.string_types):
+                metadata['date'] = get_date(settings['DEFAULT_DATE'])
+            else:
+                metadata['date'] = SafeDatetime(*settings['DEFAULT_DATE'])
     return metadata
 
 
@@ -601,20 +688,20 @@ def path_metadata(full_path, source_path, settings=None):
     if settings:
         if settings.get('DEFAULT_DATE', None) == 'fs':
             metadata['date'] = SafeDatetime.fromtimestamp(
-                os.stat(full_path).st_ctime)
+                os.stat(full_path).st_mtime)
         metadata.update(settings.get('EXTRA_PATH_METADATA', {}).get(
             source_path, {}))
     return metadata
 
 
 def parse_path_metadata(source_path, settings=None, process=None):
-    """Extract a metadata dictionary from a file's path
+    r"""Extract a metadata dictionary from a file's path
 
     >>> import pprint
     >>> settings = {
-    ...     'FILENAME_METADATA': '(?P<slug>[^.]*).*',
+    ...     'FILENAME_METADATA': r'(?P<slug>[^.]*).*',
     ...     'PATH_METADATA':
-    ...         '(?P<category>[^/]*)/(?P<date>\d{4}-\d{2}-\d{2})/.*',
+    ...         r'(?P<category>[^/]*)/(?P<date>\d{4}-\d{2}-\d{2})/.*',
     ...     }
     >>> reader = BaseReader(settings=settings)
     >>> metadata = parse_path_metadata(
@@ -636,15 +723,15 @@ def parse_path_metadata(source_path, settings=None, process=None):
                           ('PATH_METADATA', source_path)]:
             checks.append((settings.get(key, None), data))
         if settings.get('USE_FOLDER_AS_CATEGORY', None):
-            checks.insert(0, ('(?P<category>.*)', subdir))
+            checks.append(('(?P<category>.*)', subdir))
         for regexp, data in checks:
             if regexp and data:
                 match = re.match(regexp, data)
                 if match:
                     # .items() for py3k compat.
                     for k, v in match.groupdict().items():
-                        if k not in metadata:
-                            k = k.lower()  # metadata must be lowercase
+                        k = k.lower()  # metadata must be lowercase
+                        if v is not None and k not in metadata:
                             if process:
                                 v = process(k, v)
                             metadata[k] = v
